@@ -8,6 +8,7 @@
 //
 
 #include <cstdlib>
+#include <boost/regex.hpp>
 #include <boost/logic/tribool.hpp>
 #include <pion/net/HTTPParser.hpp>
 #include <pion/net/HTTPRequest.hpp>
@@ -32,11 +33,14 @@ const boost::uint32_t	HTTPParser::QUERY_VALUE_MAX = 1024 * 1024;	// 1 MB
 const boost::uint32_t	HTTPParser::COOKIE_NAME_MAX = 1024;	// 1 KB
 const boost::uint32_t	HTTPParser::COOKIE_VALUE_MAX = 1024 * 1024;	// 1 MB
 const std::size_t		HTTPParser::DEFAULT_CONTENT_MAX = 1024 * 1024;	// 1 MB
+HTTPParser::ErrorCategory *	HTTPParser::m_error_category_ptr = NULL;
+boost::once_flag			HTTPParser::m_instance_flag = BOOST_ONCE_INIT;
 
 
 // HTTPParser member functions
 
-boost::tribool HTTPParser::parse(HTTPMessage& http_msg)
+boost::tribool HTTPParser::parse(HTTPMessage& http_msg,
+	boost::system::error_code& ec)
 {
 	PION_ASSERT(! eof() );
 
@@ -56,18 +60,18 @@ boost::tribool HTTPParser::parse(HTTPMessage& http_msg)
 
 			// parsing the HTTP headers
 			case PARSE_HEADERS:
-				rc = parseHeaders(http_msg);
+				rc = parseHeaders(http_msg, ec);
 				total_bytes_parsed += m_bytes_last_read;
 				// check if we have finished parsing HTTP headers
 				if (rc == true) {
 					// finishHeaderParsing() updates m_message_parse_state
-					rc = finishHeaderParsing(http_msg);
+					rc = finishHeaderParsing(http_msg, ec);
 				}
 				break;
 
 			// parsing chunked payload content
 			case PARSE_CHUNKS:
-				rc = parseChunks(http_msg.getChunkCache());
+				rc = parseChunks(http_msg.getChunkCache(), ec);
 				total_bytes_parsed += m_bytes_last_read;
 				// check if we have finished parsing all chunks
 				if (rc == true) {
@@ -77,7 +81,7 @@ boost::tribool HTTPParser::parse(HTTPMessage& http_msg)
 
 			// parsing regular payload content with a known length
 			case PARSE_CONTENT:
-				rc = consumeContent(http_msg);
+				rc = consumeContent(http_msg, ec);
 				total_bytes_parsed += m_bytes_last_read;
 				break;
 
@@ -108,7 +112,8 @@ boost::tribool HTTPParser::parse(HTTPMessage& http_msg)
 	return rc;
 }
 
-boost::tribool HTTPParser::parseMissingData(HTTPMessage& http_msg, std::size_t len)
+boost::tribool HTTPParser::parseMissingData(HTTPMessage& http_msg,
+	std::size_t len, boost::system::error_code& ec)
 {
 	static const char MISSING_DATA_CHAR = 'X';
 	boost::tribool rc = boost::indeterminate;
@@ -120,6 +125,7 @@ boost::tribool HTTPParser::parseMissingData(HTTPMessage& http_msg, std::size_t l
 		// cannot recover from missing data while parsing HTTP headers
 		case PARSE_START:
 		case PARSE_HEADERS:
+			setError(ec, ERROR_MISSING_HEADER_DATA);
 			rc = false;
 			break;
 
@@ -144,6 +150,7 @@ boost::tribool HTTPParser::parseMissingData(HTTPMessage& http_msg, std::size_t l
 				}
 			} else {
 				// cannot recover from missing data
+				setError(ec, ERROR_MISSING_CHUNK_DATA);
 				rc = false;
 			}
 			break;
@@ -156,6 +163,7 @@ boost::tribool HTTPParser::parseMissingData(HTTPMessage& http_msg, std::size_t l
 				rc = true;
 			} else if (m_bytes_content_remaining < len) {
 				// cannot recover from missing data
+				setError(ec, ERROR_MISSING_TOO_MUCH_CONTENT);
 				rc = false;
 			} else {
 
@@ -204,7 +212,8 @@ boost::tribool HTTPParser::parseMissingData(HTTPMessage& http_msg, std::size_t l
 	return rc;
 }
 
-boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
+boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg,
+	boost::system::error_code& ec)
 {
 	//
 	// note that boost::tribool may have one of THREE states:
@@ -217,12 +226,17 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 	m_bytes_last_read = 0;
 	while (m_read_ptr < m_read_end_ptr) {
 
+		if (m_save_raw_headers)
+			m_raw_headers += *m_read_ptr;
+		
 		switch (m_headers_parse_state) {
 		case PARSE_METHOD_START:
 			// we have not yet started parsing the HTTP method string
 			if (*m_read_ptr != ' ' && *m_read_ptr!='\r' && *m_read_ptr!='\n') {	// ignore leading whitespace
-				if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr))
+				if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+					setError(ec, ERROR_METHOD_CHAR);
 					return false;
+				}
 				m_headers_parse_state = PARSE_METHOD;
 				m_method.erase();
 				m_method.push_back(*m_read_ptr);
@@ -235,8 +249,10 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				m_resource.erase();
 				m_headers_parse_state = PARSE_URI_STEM;
 			} else if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+				setError(ec, ERROR_METHOD_CHAR);
 				return false;
 			} else if (m_method.size() >= METHOD_MAX) {
+				setError(ec, ERROR_METHOD_SIZE);
 				return false;
 			} else {
 				m_method.push_back(*m_read_ptr);
@@ -250,9 +266,19 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			} else if (*m_read_ptr == '?') {
 				m_query_string.erase();
 				m_headers_parse_state = PARSE_URI_QUERY;
+			} else if (*m_read_ptr == '\r') {
+				http_msg.setVersionMajor(0);
+				http_msg.setVersionMinor(0);
+				m_headers_parse_state = PARSE_EXPECTING_NEWLINE;
+			} else if (*m_read_ptr == '\n') {
+				http_msg.setVersionMajor(0);
+				http_msg.setVersionMinor(0);
+				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else if (isControl(*m_read_ptr)) {
+				setError(ec, ERROR_URI_CHAR);
 				return false;
 			} else if (m_resource.size() >= RESOURCE_MAX) {
+				setError(ec, ERROR_URI_SIZE);
 				return false;
 			} else {
 				m_resource.push_back(*m_read_ptr);
@@ -264,8 +290,10 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			if (*m_read_ptr == ' ') {
 				m_headers_parse_state = PARSE_HTTP_VERSION_H;
 			} else if (isControl(*m_read_ptr)) {
+				setError(ec, ERROR_QUERY_CHAR);
 				return false;
 			} else if (m_query_string.size() >= QUERY_STRING_MAX) {
+				setError(ec, ERROR_QUERY_SIZE);
 				return false;
 			} else {
 				m_query_string.push_back(*m_read_ptr);
@@ -274,37 +302,73 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 
 		case PARSE_HTTP_VERSION_H:
 			// parsing "HTTP"
-			if (*m_read_ptr != 'H') return false;
+			if (*m_read_ptr == '\r') {
+				// should only happen for requests (no HTTP/VERSION specified)
+				if (! m_is_request) {
+					setError(ec, ERROR_VERSION_EMPTY);
+					return false;
+				}
+				http_msg.setVersionMajor(0);
+				http_msg.setVersionMinor(0);
+				m_headers_parse_state = PARSE_EXPECTING_NEWLINE;
+			} else if (*m_read_ptr == '\n') {
+				// should only happen for requests (no HTTP/VERSION specified)
+				if (! m_is_request) {
+					setError(ec, ERROR_VERSION_EMPTY);
+					return false;
+				}
+				http_msg.setVersionMajor(0);
+				http_msg.setVersionMinor(0);
+				m_headers_parse_state = PARSE_EXPECTING_CR;
+			} else if (*m_read_ptr != 'H') {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			m_headers_parse_state = PARSE_HTTP_VERSION_T_1;
 			break;
 
 		case PARSE_HTTP_VERSION_T_1:
 			// parsing "HTTP"
-			if (*m_read_ptr != 'T') return false;
+			if (*m_read_ptr != 'T') {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			m_headers_parse_state = PARSE_HTTP_VERSION_T_2;
 			break;
 
 		case PARSE_HTTP_VERSION_T_2:
 			// parsing "HTTP"
-			if (*m_read_ptr != 'T') return false;
+			if (*m_read_ptr != 'T') {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			m_headers_parse_state = PARSE_HTTP_VERSION_P;
 			break;
 
 		case PARSE_HTTP_VERSION_P:
 			// parsing "HTTP"
-			if (*m_read_ptr != 'P') return false;
+			if (*m_read_ptr != 'P') {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			m_headers_parse_state = PARSE_HTTP_VERSION_SLASH;
 			break;
 
 		case PARSE_HTTP_VERSION_SLASH:
 			// parsing slash after "HTTP"
-			if (*m_read_ptr != '/') return false;
+			if (*m_read_ptr != '/') {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			m_headers_parse_state = PARSE_HTTP_VERSION_MAJOR_START;
 			break;
 
 		case PARSE_HTTP_VERSION_MAJOR_START:
 			// parsing the first digit of the major version number
-			if (!isDigit(*m_read_ptr)) return false;
+			if (!isDigit(*m_read_ptr)) {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			http_msg.setVersionMajor(*m_read_ptr - '0');
 			m_headers_parse_state = PARSE_HTTP_VERSION_MAJOR;
 			break;
@@ -317,13 +381,17 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				http_msg.setVersionMajor( (http_msg.getVersionMajor() * 10)
 										  + (*m_read_ptr - '0') );
 			} else {
+				setError(ec, ERROR_VERSION_CHAR);
 				return false;
 			}
 			break;
 
 		case PARSE_HTTP_VERSION_MINOR_START:
 			// parsing the first digit of the minor version number
-			if (!isDigit(*m_read_ptr)) return false;
+			if (!isDigit(*m_read_ptr)) {
+				setError(ec, ERROR_VERSION_CHAR);
+				return false;
+			}
 			http_msg.setVersionMinor(*m_read_ptr - '0');
 			m_headers_parse_state = PARSE_HTTP_VERSION_MINOR;
 			break;
@@ -331,28 +399,39 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 		case PARSE_HTTP_VERSION_MINOR:
 			// parsing the major version number (not first digit)
 			if (*m_read_ptr == ' ') {
-				// should only happen for responses
-				if (m_is_request) return false;
-				m_headers_parse_state = PARSE_STATUS_CODE_START;
+				// ignore trailing spaces after version in request
+				if (! m_is_request) {
+					m_headers_parse_state = PARSE_STATUS_CODE_START;
+				}
 			} else if (*m_read_ptr == '\r') {
 				// should only happen for requests
-				if (! m_is_request) return false;
+				if (! m_is_request) {
+					setError(ec, ERROR_STATUS_EMPTY);
+					return false;
+				}
 				m_headers_parse_state = PARSE_EXPECTING_NEWLINE;
 			} else if (*m_read_ptr == '\n') {
 				// should only happen for requests
-				if (! m_is_request) return false;
+				if (! m_is_request) {
+					setError(ec, ERROR_STATUS_EMPTY);
+					return false;
+				}
 				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else if (isDigit(*m_read_ptr)) {
 				http_msg.setVersionMinor( (http_msg.getVersionMinor() * 10)
 										  + (*m_read_ptr - '0') );
 			} else {
+				setError(ec, ERROR_VERSION_CHAR);
 				return false;
 			}
 			break;
 
 		case PARSE_STATUS_CODE_START:
 			// parsing the first digit of the response status code
-			if (!isDigit(*m_read_ptr)) return false;
+			if (!isDigit(*m_read_ptr)) {
+				setError(ec, ERROR_STATUS_CHAR);
+				return false;
+			}
 			m_status_code = (*m_read_ptr - '0');
 			m_headers_parse_state = PARSE_STATUS_CODE;
 			break;
@@ -364,7 +443,16 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				m_headers_parse_state = PARSE_STATUS_MESSAGE;
 			} else if (isDigit(*m_read_ptr)) {
 				m_status_code = ( (m_status_code * 10) + (*m_read_ptr - '0') );
+			} else if (*m_read_ptr == '\r') {
+				// recover from status message not sent
+				m_status_message.erase();
+				m_headers_parse_state = PARSE_EXPECTING_NEWLINE;
+			} else if (*m_read_ptr == '\n') {
+				// recover from status message not sent
+				m_status_message.erase();
+				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else {
+				setError(ec, ERROR_STATUS_CHAR);
 				return false;
 			}
 			break;
@@ -376,8 +464,10 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			} else if (*m_read_ptr == '\n') {
 				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else if (isControl(*m_read_ptr)) {
+				setError(ec, ERROR_STATUS_CHAR);
 				return false;
 			} else if (m_status_message.size() >= STATUS_MESSAGE_MAX) {
+				setError(ec, ERROR_STATUS_CHAR);
 				return false;
 			} else {
 				m_status_message.push_back(*m_read_ptr);
@@ -399,6 +489,7 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			} else if (*m_read_ptr == '\t' || *m_read_ptr == ' ') {
 				m_headers_parse_state = PARSE_HEADER_WHITESPACE;
 			} else if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+				setError(ec, ERROR_HEADER_CHAR);
 				return false;
 			} else {
 				// assume it is the first character for the name of a header
@@ -423,6 +514,7 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			} else if (*m_read_ptr == '\t' || *m_read_ptr == ' ') {
 				m_headers_parse_state = PARSE_HEADER_WHITESPACE;
 			} else if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+				setError(ec, ERROR_HEADER_CHAR);
 				return false;
 			} else {
 				// assume it is the first character for the name of a header
@@ -440,6 +532,7 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else if (*m_read_ptr != '\t' && *m_read_ptr != ' ') {
 				if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr))
+					setError(ec, ERROR_HEADER_CHAR);
 					return false;
 				// assume it is the first character for the name of a header
 				m_header_name.erase();
@@ -457,6 +550,7 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			} else if (*m_read_ptr == '\t' || *m_read_ptr == ' ') {
 				m_headers_parse_state = PARSE_HEADER_WHITESPACE;
 			} else if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+				setError(ec, ERROR_HEADER_CHAR);
 				return false;
 			} else {
 				// first character for the name of a header
@@ -472,8 +566,10 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				m_header_value.erase();
 				m_headers_parse_state = PARSE_SPACE_BEFORE_HEADER_VALUE;
 			} else if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+				setError(ec, ERROR_HEADER_CHAR);
 				return false;
 			} else if (m_header_name.size() >= HEADER_NAME_MAX) {
+				setError(ec, ERROR_HEADER_NAME_SIZE);
 				return false;
 			} else {
 				// character (not first) for the name of a header
@@ -492,6 +588,7 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				http_msg.addHeader(m_header_name, m_header_value);
 				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else if (!isChar(*m_read_ptr) || isControl(*m_read_ptr) || isSpecial(*m_read_ptr)) {
+				setError(ec, ERROR_HEADER_CHAR);
 				return false;
 			} else {
 				// assume it is the first character for the value of a header
@@ -509,8 +606,10 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 				http_msg.addHeader(m_header_name, m_header_value);
 				m_headers_parse_state = PARSE_EXPECTING_CR;
 			} else if (isControl(*m_read_ptr)) {
+				setError(ec, ERROR_HEADER_CHAR);
 				return false;
 			} else if (m_header_value.size() >= HEADER_VALUE_MAX) {
+				setError(ec, ERROR_HEADER_VALUE_SIZE);
 				return false;
 			} else {
 				// character (not first) for the value of a header
@@ -531,9 +630,6 @@ boost::tribool HTTPParser::parseHeaders(HTTPMessage& http_msg)
 			return true;
 		}
 		
-		if (m_save_raw_headers)
-			m_raw_headers += *m_read_ptr;
-		
 		++m_read_ptr;
 	}
 
@@ -558,8 +654,7 @@ void HTTPParser::updateMessageWithHeaderData(HTTPMessage& http_msg) const
 			if (! parseURLEncoded(http_request.getQueryParams(),
 								  m_query_string.c_str(),
 								  m_query_string.size())) 
-				PION_LOG_WARN(m_logger, "Request query string parsing failed (URI): \""
-					<< m_query_string << "\"");
+				PION_LOG_WARN(m_logger, "Request query string parsing failed (URI)");
 		}
 
 		// parse "Cookie" headers in request
@@ -597,7 +692,8 @@ void HTTPParser::updateMessageWithHeaderData(HTTPMessage& http_msg) const
 	}
 }
 
-boost::tribool HTTPParser::finishHeaderParsing(HTTPMessage& http_msg)
+boost::tribool HTTPParser::finishHeaderParsing(HTTPMessage& http_msg,
+	boost::system::error_code& ec)
 {
 	boost::tribool rc = boost::indeterminate;
 
@@ -610,6 +706,10 @@ boost::tribool HTTPParser::finishHeaderParsing(HTTPMessage& http_msg)
 
 		// content is encoded using chunks
 		m_message_parse_state = PARSE_CHUNKS;
+		
+		// return true if parsing headers only
+		if (m_parse_headers_only)
+			rc = true;
 
 	} else if (http_msg.isContentLengthImplied()) {
 
@@ -627,6 +727,7 @@ boost::tribool HTTPParser::finishHeaderParsing(HTTPMessage& http_msg)
 				http_msg.updateContentLengthUsingHeader();
 			} catch (...) {
 				PION_LOG_ERROR(m_logger, "Unable to update content length");
+				setError(ec, ERROR_INVALID_CONTENT_LENGTH);
 				return false;
 			}
 
@@ -641,6 +742,10 @@ boost::tribool HTTPParser::finishHeaderParsing(HTTPMessage& http_msg)
 				// check if content-length exceeds maximum allowed
 				if (m_bytes_content_remaining > m_max_content_length)
 					http_msg.setContentLength(m_max_content_length);
+
+				// return true if parsing headers only
+				if (m_parse_headers_only)
+					rc = true;
 			}
 
 		} else {
@@ -654,6 +759,10 @@ boost::tribool HTTPParser::finishHeaderParsing(HTTPMessage& http_msg)
 
 				// continue reading content until there is no more data
 				m_message_parse_state = PARSE_CONTENT_NO_LENGTH;
+
+				// return true if parsing headers only
+				if (m_parse_headers_only)
+					rc = true;
 			} else {
 				m_message_parse_state = PARSE_END;
 				rc = true;
@@ -687,8 +796,7 @@ bool HTTPParser::parseURLEncoded(HTTPTypes::QueryParams& dict,
 		case QUERY_PARSE_NAME:
 			// parsing query name
 			if (*ptr == '=') {
-				// end of name found
-				if (query_name.empty()) return false;
+				// end of name found (OK if empty)
 				parse_state = QUERY_PARSE_VALUE;
 			} else if (*ptr == '&') {
 				// if query name is empty, just skip it (i.e. "&&")
@@ -697,8 +805,8 @@ bool HTTPParser::parseURLEncoded(HTTPTypes::QueryParams& dict,
 					dict.insert( std::make_pair(query_name, query_value) );
 					query_name.erase();
 				}
-			} else if (*ptr == '\r' || *ptr == '\n') {
-				// ignore linefeeds and carriage returns (normally within POST content)
+			} else if (*ptr == '\r' || *ptr == '\n' || *ptr == '\t') {
+				// ignore linefeeds, carriage return and tabs (normally within POST content)
 			} else if (isControl(*ptr) || query_name.size() >= QUERY_NAME_MAX) {
 				// control character detected, or max sized exceeded
 				return false;
@@ -712,12 +820,14 @@ bool HTTPParser::parseURLEncoded(HTTPTypes::QueryParams& dict,
 			// parsing query value
 			if (*ptr == '&') {
 				// end of value found (OK if empty)
-				dict.insert( std::make_pair(query_name, query_value) );
-				query_name.erase();
+				if (! query_name.empty()) {
+					dict.insert( std::make_pair(query_name, query_value) );
+					query_name.erase();
+				}
 				query_value.erase();
 				parse_state = QUERY_PARSE_NAME;
-			} else if (*ptr == '\r' || *ptr == '\n') {
-				// ignore linefeeds and carriage returns (normally within POST content)
+			} else if (*ptr == '\r' || *ptr == '\n' || *ptr == '\t') {
+				// ignore linefeeds, carriage return and tabs (normally within POST content)
 			} else if (isControl(*ptr) || query_value.size() >= QUERY_VALUE_MAX) {
 				// control character detected, or max sized exceeded
 				return false;
@@ -766,8 +876,7 @@ bool HTTPParser::parseCookieHeader(HTTPTypes::CookieParams& dict,
 		case COOKIE_PARSE_NAME:
 			// parsing cookie name
 			if (*ptr == '=') {
-				// end of name found
-				if (cookie_name.empty()) return false;
+				// end of name found (OK if empty)
 				value_quote_character = '\0';
 				parse_state = COOKIE_PARSE_VALUE;
 			} else if (*ptr == ';' || *ptr == ',') {
@@ -810,7 +919,7 @@ bool HTTPParser::parseCookieHeader(HTTPTypes::CookieParams& dict,
 						// assume character is part of the (unquoted) value
 						cookie_value.push_back(*ptr);
 					}
-				} else if (*ptr != ' ') {	// ignore unquoted whitespace
+				} else if (*ptr != ' ' || !cookie_value.empty()) {	// ignore leading unquoted whitespace
 					// check if control character detected, or max sized exceeded
 					if (isControl(*ptr) || cookie_value.size() >= COOKIE_VALUE_MAX)
 						return false;
@@ -853,7 +962,8 @@ bool HTTPParser::parseCookieHeader(HTTPTypes::CookieParams& dict,
 	return true;
 }
 
-boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
+boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache,
+	boost::system::error_code& ec)
 {
 	//
 	// note that boost::tribool may have one of THREE states:
@@ -878,6 +988,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 				// but we'll be flexible, since there's no ambiguity.
 				break;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -892,6 +1003,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 				// but we'll be flexible, since there's no ambiguity.
 				m_chunked_content_parse_state = PARSE_EXPECTING_CR_AFTER_CHUNK_SIZE;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -904,6 +1016,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 				// but we'll be flexible, since there's no ambiguity.
 				break;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -920,6 +1033,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 					m_chunked_content_parse_state = PARSE_CHUNK;
 				}
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -940,6 +1054,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 			if (*m_read_ptr == '\x0D') {
 				m_chunked_content_parse_state = PARSE_EXPECTING_LF_AFTER_CHUNK;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -949,6 +1064,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 			if (*m_read_ptr == '\x0A') {
 				m_chunked_content_parse_state = PARSE_CHUNK_SIZE_START;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -958,6 +1074,7 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 			if (*m_read_ptr == '\x0D') {
 				m_chunked_content_parse_state = PARSE_EXPECTING_FINAL_LF_AFTER_LAST_CHUNK;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 			break;
@@ -968,9 +1085,11 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 				++m_read_ptr;
 				m_bytes_last_read = (m_read_ptr - read_start_ptr);
 				m_bytes_total_read += m_bytes_last_read;
+				m_bytes_content_read += m_bytes_last_read;
 				PION_LOG_DEBUG(m_logger, "Parsed " << m_bytes_last_read << " chunked payload content bytes; chunked content complete.");
 				return true;
 			} else {
+				setError(ec, ERROR_CHUNK_CHAR);
 				return false;
 			}
 		}
@@ -984,7 +1103,8 @@ boost::tribool HTTPParser::parseChunks(HTTPMessage::ChunkCache& chunk_cache)
 	return boost::indeterminate;
 }
 
-boost::tribool HTTPParser::consumeContent(HTTPMessage& http_msg)
+boost::tribool HTTPParser::consumeContent(HTTPMessage& http_msg,
+	boost::system::error_code& ec)
 {
 	size_t content_bytes_to_read;
 	size_t content_bytes_available = bytes_available();
@@ -1065,7 +1185,7 @@ void HTTPParser::finish(HTTPMessage& http_msg) const
 		http_msg.setContentLength(getContentBytesRead());
 		break;
 	case PARSE_CHUNKS:
-		http_msg.setIsValid(false);
+		http_msg.setIsValid(m_chunked_content_parse_state==PARSE_CHUNK_SIZE_START);
 		http_msg.concatenateChunks();
 		break;
 	case PARSE_CONTENT_NO_LENGTH:
@@ -1088,8 +1208,7 @@ void HTTPParser::finish(HTTPMessage& http_msg) const
 			if (! parseURLEncoded(http_request.getQueryParams(),
 								  http_request.getContent(),
 								  http_request.getContentLength())) 
-				PION_LOG_WARN(m_logger, "Request query string parsing failed (POST content): \""
-					<< http_request.getContent() << "\"");
+				PION_LOG_WARN(m_logger, "Request query string parsing failed (POST content)");
 		}
 	}
 }
@@ -1101,13 +1220,57 @@ void HTTPParser::computeMsgStatus(HTTPMessage& http_msg, bool msg_parsed_ok )
 	if(http_msg.hasMissingPackets()) {
 		st = http_msg.hasDataAfterMissingPackets() ?
 						HTTPMessage::STATUS_PARTIAL : HTTPMessage::STATUS_TRUNCATED;
-		http_msg.setStatus(st);
 	} else {
 		st = msg_parsed_ok ? HTTPMessage::STATUS_OK : HTTPMessage::STATUS_TRUNCATED;
 	}
 
 	http_msg.setStatus(st);
 }
+
+void HTTPParser::createErrorCategory(void)
+{
+	static ErrorCategory UNIQUE_ERROR_CATEGORY;
+	m_error_category_ptr = &UNIQUE_ERROR_CATEGORY;
+}
+
+bool HTTPParser::parseForwardedFor(const std::string& header, std::string& public_ip)
+{
+	// static regex's used to check for ipv4 address
+	static const boost::regex IPV4_ADDR_RX("[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}");
+
+	/// static regex used to check for private/local networks:
+	/// 10.*
+	/// 127.*
+	/// 192.168.*
+	/// 172.16-31.*
+	static const boost::regex PRIVATE_NET_RX("(10\\.[0-9]{1,3}|127\\.[0-9]{1,3}|192\\.168|172\\.1[6-9]|172\\.2[0-9]|172\\.3[0-1])\\.[0-9]{1,3}\\.[0-9]{1,3}");
+
+	// sanity check
+	if (header.empty())
+		return false;
+
+	// local variables re-used by while loop
+	boost::match_results<std::string::const_iterator> m;
+	std::string::const_iterator start_it = header.begin();
+
+	// search for next ip address within the header
+	while (boost::regex_search(start_it, header.end(), m, IPV4_ADDR_RX)) {
+		// get ip that matched
+		std::string ip_str(m[0].first, m[0].second);
+		// check if public network ip address
+		if (! boost::regex_match(ip_str, PRIVATE_NET_RX) ) {
+			// match found!
+			public_ip = ip_str;
+			return true;
+		}
+		// update search starting position
+		start_it = m[0].second;
+	}
+
+	// no matches found
+	return false;
+}
+
 }	// end namespace net
 }	// end namespace pion
 
